/*
 * simple UDP forwarder
 *
 * Copyright (c) by TWISTI in 3/2000
 *
 */

#include <stdio.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <string.h>
#include <stdlib.h>
#include <errno.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <ctype.h>
#include <wait.h>
#include <signal.h>


#define CONF_FILE   "/etc/uinetd.conf"
char conf_file[256];


int in_sockfd = 0;
int out_sockfd = 0;
FILE *file;
char *progname;
int verbose = 0;



/*
 * clean up the ports and files (child exit function)
 */
void clean_exit (int sig) 
{
  if (fclose (file) == EOF) {
      (void) fprintf (stdout, "%s: error closing config file: %s\n", progname, strerror (errno));
  }

  if (close (in_sockfd) < 0) {
      (void) fprintf (stdout, "%s: error closing input socket: %s\n", progname, strerror (errno));
  }

  if (close (out_sockfd) < 0) {
      (void) fprintf (stdout, "%s: error closing output socket: %s\n", progname, strerror (errno));
  }

  if (verbose > 0) {
      (void) printf ("child exiting cleanly!\n");
  }
  exit (0);
}



/*
 * dummy exit function for parent process
 */
void dummy_exit (int sig)
{
  if (fclose (file) == EOF) {
      (void) fprintf (stdout, "%s: error closing config file: %s\n", progname, strerror (errno));
  }
  
  /* and now return to get the childs */
}



/*
 * print usage informations
 */
void usage ()
{
  (void) fprintf (stdout, "uinetd v0.1 - UDP redirector daemon (c) 1999 by TWISTI (twisti@gmx.net)\n"
                          "Usage: %s [options]\n\n"
                          "Options:\n"
                          "    -h        display this help\n"
                          "    -v        be verbose, twice for even more verbose\n"
                          "    -f FILE   read configuration from FILE\n", progname);
}



int main (int argc, char **argv) 
{
  static int size;
  unsigned char buf[3924];		/* i think this is enough (mtu of loopback device) */
  struct sockaddr_in in_host;
  struct sockaddr_in out_host;
  struct sockaddr_in from;
  static int fromlen;
  char line[256];
  char source_addr[15];
  unsigned int source_port;
  char dest_addr[15];
  unsigned int dest_port;
  int arg;
  char proto_string[3];
  struct protoent *proto;


  progname = argv[0];			/* set program name for functions */
  (void) strncpy (conf_file, CONF_FILE, strlen (CONF_FILE) + 1);

  /*
   * get those args
   */
  while ((arg = getopt (argc, argv, "vhf:")) != EOF) {
      switch (arg) {
          case 'v': verbose++;
                    break;
          case 'h': usage ();
                    exit (0);
                    break;
          case 'f': strncpy (conf_file, optarg, strlen (optarg) + 1);	/* also copy \0 */
                    break;
      }
  }
          

  /*
   * open config file
   */
  if ((file = fopen (conf_file, "r+")) == NULL) {
      (void) fprintf (stdout, "%s: error open config file `%s': %s\n", progname, conf_file, strerror (errno));
      exit (-1);
  }

  while (fgets(line, 255, file) !=  NULL) {
      if (line[0] == '\n' || line[0] == '#' || line[0] == ' ')
          continue;

      if (sscanf (line, "%s %s %d %s %d", proto_string, source_addr, &source_port, dest_addr, &dest_port) == EOF) {
          perror ("sscanf() line from config file");
      }

      if (verbose > 0) {
          (void) printf ("from: %s/%d   to: %s/%d   proto: %s\n", source_addr, source_port, dest_addr, dest_port, proto_string);
      }

      if (fork () == 0) {

          signal (SIGHUP, clean_exit);		/* catch the signals for childs and terminate cleanly */
          signal (SIGINT, clean_exit);
          signal (SIGQUIT, clean_exit);
          signal (SIGTERM, clean_exit);

          /*
           * get protocol
           */
          if ((proto = getprotobyname (proto_string)) == NULL) {
              (void) fprintf (stdout, "%s: getprotobyname() error, maybe unknown protocol: %s\n", progname, proto_string);
              exit (-1);
          }

          /*
           * create the sockets
           */
          if (verbose > 0) {
              (void) printf ("get input socket.\n");
          }

          switch (proto->p_proto) {
              case 17: if ((in_sockfd = socket (AF_INET, SOCK_DGRAM, IPPROTO_UDP)) < 0) {	/* UDP */
                              perror ("getting udp input socket");
                              exit (-1);
                       }
                       break;
              default: (void) fprintf (stdout, "%s: protocol type not supported: %s\n", progname, proto->p_name);
                       exit (-1);
                       break;
          }

          if (verbose > 0) {
              (void) printf ("input sockfd: %d\n", in_sockfd);
          }


          if (verbose > 0) {
              (void) printf ("get output socket.\n");
          }

          switch (proto->p_proto) {
              case 17: if ((out_sockfd = socket (AF_INET, SOCK_DGRAM, IPPROTO_UDP)) < 0) {
                           perror ("getting udp output socket");
                           exit (-1);
                       }
                       break;
              default: (void) fprintf (stdout, "%s: protocol type not supported: %s\n", progname, proto->p_name);
                       exit (-1);
                       break;
          }

          if (verbose > 0) {
              (void) printf ("output sockfd: %d\n", out_sockfd);
          }


          /*
           * connecting to the port faceing to the internet
           */
          bzero (&in_host, sizeof (in_host));
          in_host.sin_family = AF_INET;
          in_host.sin_port = htons (source_port);
          in_host.sin_addr.s_addr = INADDR_ANY;
          bzero (&in_host.sin_zero, 8);

          if (verbose > 0) { 
              (void) printf ("bind() to input socket: %s/%d\n", source_addr, source_port);
          }
          if (bind (in_sockfd, (struct sockaddr *) &in_host, sizeof (struct sockaddr)) < 0) {
              perror ("bind()'ing input socket");
              exit (-1);
          }


          /*
           * connecting to the port facing inside the masqueraded network
           */
          bzero (&out_host, sizeof (out_host));
          out_host.sin_family = AF_INET;
          out_host.sin_port = htons (dest_port);
          inet_aton (dest_addr, &out_host.sin_addr);		/* converts nums and dots to inetaddr */
          bzero (&out_host.sin_zero, 8);

          if (verbose > 0) {
              (void) printf ("connect() to output socket: %s/%d\n", dest_addr, dest_port);
          }
          if (connect (out_sockfd, (struct sockaddr *) &out_host, sizeof (struct sockaddr)) < 0) {
              perror ("connect()'ing output socket");
              exit (-1);
          }


          while (1) {
              /*
               * receive data from the input port
               */
              if (verbose > 1) {
                  (void) printf ("recvfrom() input socket.\n");
              }
              if ((size = recvfrom (in_sockfd, &buf, sizeof (buf), 0, (struct sockaddr *) &from, &fromlen)) < 0) {
                  perror ("recvfrom() input socket");
                  exit (-1);
              } 

              if (verbose > 1) {
                  (void) printf ("got packet: %d\n", size);
              }

              if (verbose > 1) {
                  (void) printf ("from host: %s\n", inet_ntoa (from.sin_addr));
              }

              /*
               * send to data port of destination server
               */
              if (verbose > 0) {
                  (void) printf ("sendto() output socket.\n");
              }
              if ((size = sendto (out_sockfd, &buf, size, 0, (struct sockaddr *) &out_host, sizeof (out_host))) < 0) {
                  perror ("sendto() output socket");
              }
              else {
                  if (verbose > 1) {
                      (void) printf ("sent packet: %d\n", size);
                  }
              }
          }
      } 
  }

  signal (SIGHUP, dummy_exit);		/* catch the signals for the parent process */
  signal (SIGINT, dummy_exit);
  signal (SIGQUIT, dummy_exit);
  signal (SIGTERM, dummy_exit);

  if (verbose > 0) {
      (void) printf ("wait()'ing for childs\n");
  }
  if (wait (NULL) < 0) {
      (void) printf ("%s: wait() error: %s\n", progname, strerror (errno));
      exit (-1);
  }

  if (verbose > 0) {
      (void) printf ("got childs!\n");
  }
  return 0;
}

